#ifdef SCREEN_i2c

#include <brzo_i2c.h> ///BRZO_I2C_DISABLE_INTERRUPTS 1 must be set to 0 there!

#if F_CPU == 160000000L
#define BRZO_I2C_SPEED 1000
#else
#define BRZO_I2C_SPEED 800
#endif

#endif



#define VIEW_WDT	    128
#define VIEW_HGT	    64

enum {
  DM_OR = 0,
  DM_EOR,
  DM_AND
};

#define CMD_CHARGEPUMP              0x8D
#define CMD_COMSCANDEC              0xC8
#define CMD_COMSCANINC              0xC0
#define CMD_DISPLAYALLON_RESUME     0xA4
#define CMD_DISPLAYOFF              0xAE
#define CMD_DISPLAYON               0xAF
#define CMD_EXTERNALVCC             0x01
#define CMD_INVERTDISPLAY           0xA7
#define CMD_MEMORYMODE              0x20
#define CMD_NORMALDISPLAY           0xA6
#define CMD_SEGREMAP                0xA0
#define CMD_SETCOMPINS              0xDA
#define CMD_SETCONTRAST             0x81
#define CMD_SETDISPLAYCLOCKDIV      0xD5
#define CMD_SETDISPLAYOFFSET        0xD3
#define CMD_SETHIGHCOLUMN           0x10
#define CMD_SETLOWCOLUMN            0x00
#define CMD_SETMULTIPLEX            0xA8
#define CMD_SETPRECHARGE            0xD9
#define CMD_SETSTARTLINE            0x40
#define CMD_SETVCOMDETECT           0xDB
#define CMD_SWITCHCAPVCC            0x02


#ifdef SCREEN_i2c

#define PACKET_SIZE   VIEW_WDT    //bytes per transfer, up to VIEW_WDT
uint8_t dv_i2c_adr;

#endif

#ifdef SCREEN_SPI

#define PACKET_SIZE   VIEW_WDT

uint8_t dv_spi_cs;
uint8_t dv_spi_dc;

#ifdef SCREEN_SPI_FAST
SPISettings dv_spi_settings(16000000, MSBFIRST, SPI_MODE0);  //8 MHz by datasheet, but works at 16 MHz too
#endif

#endif

uint8_t dv_mode;

uint8_t dv_raster[VIEW_HGT / 8 * VIEW_WDT] __attribute__((aligned(4))); //bytes oriented vertically into 8 128-byte strips
uint8_t dv_change[VIEW_HGT / 8];
uint8_t dv_mask[2];

const uint8_t* dv_font_data;
const uint8_t* dv_font_width;

bool dv_enable = false;
bool dv_enable_prev = true;



void display_cmd(uint8_t cmd)
{
#ifdef SCREEN_i2c

  static uint8_t cmd_buf[2];

  cmd_buf[0] = 0x80;
  cmd_buf[1] = cmd;

  brzo_i2c_start_transaction(dv_i2c_adr, BRZO_I2C_SPEED);
  brzo_i2c_write(cmd_buf, sizeof(cmd_buf), true);
  brzo_i2c_end_transaction();

#endif

#ifdef SCREEN_SPI

  //digitalWrite(dv_spi_cs, HIGH);
  digitalWrite(dv_spi_dc, LOW);
  digitalWrite(dv_spi_cs, LOW);

#ifdef SCREEN_SPI_FAST
  SPI.beginTransaction(dv_spi_settings);
  SPI.transfer(cmd);
  SPI.endTransaction();
#else
  SPI.transfer(cmd);
#endif

  digitalWrite(dv_spi_cs, HIGH);

#endif
}


void display_clear()
{
  memset(dv_raster, 0, sizeof(dv_raster));
  memset(dv_change, 1, sizeof(dv_change));
}



void display_mask(uint8_t n)
{
  const uint8_t mask[] = {
    0xff, 0xff,
    0xff, 0xaa,
    0x55, 0xaa,
    0x55, 0x00
  };

  if (n >= 0 && n < sizeof(mask) / 2)
  {
    dv_mask[0] = mask[n * 2 + 0];
    dv_mask[1] = mask[n * 2 + 1];
  }
}



void display_mode(uint8_t mode)
{
  dv_mode = mode;
}



void display_brightness(int br)
{
  if (br < 0) br = 0;
  if (br > 100) br = 100;

  br = 255 * br / 100;

  uint8_t co = br * 1.171;

  if (br >= 128) co -= 43;

  display_cmd(CMD_SETPRECHARGE);
  display_cmd(br ? 241 : 0);
  display_cmd(CMD_SETCONTRAST);
  display_cmd(co);
  display_cmd(CMD_SETVCOMDETECT);
  display_cmd(co / 8);
  display_cmd(CMD_DISPLAYALLON_RESUME);
  display_cmd(CMD_NORMALDISPLAY);
  display_cmd(dv_enable ? CMD_DISPLAYON : CMD_DISPLAYOFF);
}



void display_enable(bool ena)
{
  dv_enable = ena;
}



bool display_is_active()
{
  return dv_enable;
}



void display_invert(bool inv)
{
  display_cmd(!inv ? CMD_NORMALDISPLAY : CMD_INVERTDISPLAY);
}



void display_orient(bool vflip, bool hflip)
{
  display_cmd(CMD_SEGREMAP);
  display_cmd(CMD_COMSCANINC);

  if (vflip)
  {
    display_cmd(CMD_SEGREMAP | 0x01);
    display_cmd(CMD_COMSCANDEC);
  }

  if (hflip)
  {
    display_cmd(CMD_SEGREMAP);
    display_cmd(CMD_COMSCANDEC);
  }
}



void display_char(int x, int y, char c)
{
  if (c >= 128) return;

  int cy = y / 8;

  int ps = c * 8;
  int pd = x + cy * VIEW_WDT;

  int cy1 = (cy >= 0 && cy < VIEW_HGT / 8) ? 1 : 0;

  ++cy;

  int cy2 = (cy >= 0 && cy < VIEW_HGT / 8) ? 1 : 0;

  if (cy1) dv_change[cy - 1] = 1;
  if (cy2) dv_change[cy + 0] = 1;

  int cwdt = pgm_read_byte(dv_font_width + c);

  switch (dv_mode)
  {
    case DM_OR:
      for (int i = 0; i < cwdt; ++i)
      {
        if (x >= 0 && x < VIEW_WDT)
        {
          int ln = pgm_read_byte(dv_font_data + ps) << (y & 7);

          if (cy1) dv_raster[pd] |= (ln & 255) & dv_mask[x & 1];
          if (cy2) dv_raster[pd + VIEW_WDT] |= (ln >> 8) & dv_mask[x & 1];
        }

        ++x;
        ++ps;
        ++pd;
      }
      break;

    case DM_EOR:
      for (int i = 0; i < cwdt; ++i)
      {
        if (x >= 0 && x < VIEW_WDT)
        {
          int ln = pgm_read_byte(dv_font_data + ps) << (y & 7);

          if (cy1) dv_raster[pd] ^= (ln & 255) & dv_mask[x & 1];
          if (cy2) dv_raster[pd + VIEW_WDT] ^= (ln >> 8) & dv_mask[x & 1];
        }

        ++x;
        ++ps;
        ++pd;
      }
      break;

    case DM_AND:
      for (int i = 0; i < cwdt; ++i)
      {
        if (x >= 0 && x < VIEW_WDT)
        {
          int ln = pgm_read_byte(dv_font_data + ps) << (y & 7);

          if (cy1) dv_raster[pd] &= ~((ln & 255) & dv_mask[x & 1]);
          if (cy2) dv_raster[pd + VIEW_WDT] &= ~((ln >> 8) & dv_mask[x & 1]);
        }

        ++x;
        ++ps;
        ++pd;
      }
      break;
  }
}



int display_str_u(int x, int y, const char* str, bool pgm, bool draw)
{
  char c;
  int sx = x;

  while (1)
  {
    if (!pgm)
    {
      c = *str++;
    }
    else
    {
      c = pgm_read_byte(str++);
    }

    if (!c || c == '\n') break;

    if (draw)
    {
      if ((x >= -8) && (x < VIEW_WDT)) display_char(x, y, c);
    }

    x += pgm_read_byte(dv_font_width + c) + 1;

    if (draw && x >= VIEW_WDT) break;
  }

  return x - sx;
}



void display_str(int x, int y, const char* str)
{
  display_str_u(x, y, str, false, true);
}



void display_str_P(int x, int y, PGM_P str)
{
  display_str_u(x, y, (const char*)str, true, true);
}



int display_str_len(const char* str)
{
  return display_str_u(0, 0, str, false, false);
}



int display_str_len_P(PGM_P str)
{
  return display_str_u(0, 0, (const char*)str, true, false);
}



void display_rect(int x, int y, int w, int h)
{
  if (x >= VIEW_WDT || (x + w) < 0) return;
  if (y >= VIEW_HGT || (y + h) < 0) return;

  if (x < 0)
  {
    w += x;
    x = 0;
  }

  if (x + w > VIEW_WDT)
  {
    w = VIEW_WDT - x;
  }

  if (y < 0)
  {
    h += y;
    y = 0;
  }

  if (y + h > VIEW_HGT)
  {
    h = VIEW_HGT - y;
  }

  uint8_t etop = 0, ebtm = 0;

  for (int i = 0; i < (h < 8 ? h : 8); ++i) etop |= (1 << (i + (y & 7)));
  for (int i = 0; i < (((y + h) & 7) > 0 ? ((y + h) & 7) : 8); ++i) ebtm |= (1 << i);

  int cy = y / 8;

  uint8_t ptn = etop;

  h = (h + 7 + (y & 7)) / 8;

  while (h > 0)
  {
    if (cy >= 0 && cy < VIEW_HGT / 8)
    {
      dv_change[cy] = 1;

      int pd = cy * VIEW_WDT + x;

      switch (dv_mode)
      {
        case DM_OR:
          for (int i = 0; i < w; ++i) dv_raster[pd++] |= ptn & dv_mask[(x + i) & 1];
          break;

        case DM_EOR:
          for (int i = 0; i < w; ++i) dv_raster[pd++] ^= ptn & dv_mask[(x + i) & 1];
          break;

        case DM_AND:
          for (int i = 0; i < w; ++i) dv_raster[pd++] &= ~(ptn & dv_mask[(x + i) & 1]);
          break;
      }
    }

    ++cy;
    --h;
    if (h == 1) ptn = ebtm; else ptn = 0xff;
  }
}



void display_update()
{
  if (dv_enable)
  {
    if (pad_is_locked())
    {
      display_mode(DM_AND);
      display_mask(1);
      display_rect(0, 0, VIEW_WDT, VIEW_HGT);
      display_mask(0);
      display_rect(64 - 32, 32 - 8, 64, 16);
      display_mode(DM_OR);
      display_str_P(64 - 20, 32 - 3, "Locked");
    }

    for (int y = 0; y < VIEW_HGT / 8; ++y)
    {
      if (!dv_change[y]) continue;

      dv_change[y] = 0;

      int xoff = 2;

      display_cmd(0xb0 + y);
      display_cmd(CMD_SETLOWCOLUMN + (xoff & 0x0f));
      display_cmd(CMD_SETHIGHCOLUMN + ((xoff >> 4) & 0x0f));

#ifdef SCREEN_i2c

      brzo_i2c_start_transaction(dv_i2c_adr, BRZO_I2C_SPEED);

      bool re = true;

      for (int x = 0; x < VIEW_WDT / PACKET_SIZE; ++x)
      {
        uint8_t send_buf[1 + PACKET_SIZE];
        
        send_buf[0] = 0x40;
        
        memcpy(&send_buf[1], &dv_raster[y * VIEW_WDT + x * PACKET_SIZE], PACKET_SIZE);

        brzo_i2c_write(send_buf, sizeof(send_buf), re);

        re = false;

        yield();
      }

      brzo_i2c_end_transaction();
#endif

#ifdef SCREEN_SPI

      //digitalWrite(dv_spi_cs, HIGH);
      digitalWrite(dv_spi_dc, HIGH); //data mode
      digitalWrite(dv_spi_cs, LOW);

#ifdef SCREEN_SPI_FAST

      SPI.beginTransaction(dv_spi_settings); 
      for (int x = 0; x < VIEW_WDT / PACKET_SIZE; ++x)
      {
        int ptr = y * VIEW_WDT + x * PACKET_SIZE;
        SPI.writeBytes(&dv_raster[ptr], PACKET_SIZE);
        yield();
      }
      SPI.endTransaction();
 
#else
 
      for (int x = 0; x < VIEW_WDT / PACKET_SIZE; ++x)
      {
        int ptr = y * VIEW_WDT + x * PACKET_SIZE;
        for(int i = 0; i < PACKET_SIZE; ++i) SPI.transfer(dv_raster[ptr++]);
        yield();
      }

#endif

      digitalWrite(dv_spi_cs, HIGH);
      
#endif
    }
  }

  if (dv_enable != dv_enable_prev)
  {
    dv_enable_prev = dv_enable;

    display_cmd(dv_enable ? CMD_DISPLAYON : CMD_DISPLAYOFF);
  }
}



void display_init_std(const unsigned char* font_data, const unsigned char* font_width)
{
  dv_font_data = font_data;
  dv_font_width = font_width;

  display_cmd(CMD_DISPLAYOFF);
  display_cmd(CMD_SETDISPLAYCLOCKDIV);
  display_cmd(0xF0);
  display_cmd(CMD_SETMULTIPLEX);
  display_cmd(VIEW_HGT - 1);
  display_cmd(CMD_SETDISPLAYOFFSET);
  display_cmd(0x00);
  display_cmd(CMD_SETSTARTLINE);
  display_cmd(CMD_CHARGEPUMP);
  display_cmd(0x14);
  display_cmd(CMD_MEMORYMODE);
  display_cmd(0x00);
  display_cmd(CMD_SEGREMAP);
  display_cmd(CMD_COMSCANINC);
  display_cmd(CMD_SETCOMPINS);
  display_cmd(VIEW_HGT == 64 ? 0x12 : 0x02);
  display_cmd(CMD_SETCONTRAST);
  display_cmd(VIEW_HGT == 64 ? 0xCF : 0x8F);
  display_cmd(CMD_SETPRECHARGE);
  display_cmd(0xF1);
  display_cmd(CMD_SETVCOMDETECT);
  display_cmd(0x40);
  display_cmd(CMD_DISPLAYALLON_RESUME);
  display_cmd(CMD_NORMALDISPLAY);
  display_cmd(0x2e);

  display_mask(0);
  display_mode(DM_OR);

  display_enable(true);
  display_brightness(255);
  display_invert(false);
  display_orient(true, false);

  display_clear();
  display_update(); //performs CMD_DISPLAYON
}



#ifdef SCREEN_i2c

void display_init(int adr, int scl, int sda, const unsigned char* font_data, const unsigned char* font_width)
{
  dv_i2c_adr = adr;

  brzo_i2c_setup(sda, scl, 0);

  display_init_std(font_data, font_width);
}

#endif

#ifdef SCREEN_SPI

void display_init(int pin_cs, int pin_dc, int pin_res, const unsigned char* font_data, const unsigned char* font_width)
{
  dv_spi_cs = pin_cs;
  dv_spi_dc = pin_dc;

  pinMode(dv_spi_cs, OUTPUT);
  pinMode(dv_spi_dc, OUTPUT);
  pinMode(pin_res, OUTPUT);

  digitalWrite(dv_spi_cs, HIGH);
  digitalWrite(dv_spi_dc, HIGH);

  digitalWrite(pin_res, HIGH);
  delay(1);
  digitalWrite(pin_res, LOW);
  delay(10);
  digitalWrite(pin_res, HIGH);

  display_init_std(font_data, font_width);
}

#endif
