#define CFG_FILENAME  "config.cfg"

struct {
  uint8_t ay_layout;
  uint32_t ay_clock;
  uint16_t ay_fps;
  uint16_t scr_timeout;
  int8_t scr_bright;
  uint8_t scr_lock;
  int16_t dir_cur;
  int16_t dir_cur_prev;
  uint8_t dir_mode;
  uint8_t file_mode;
  int8_t cfg_cur;
  uint8_t sound_vol;
  char active_dir[MAX_PATH];
  char prev_dir[MAX_PATH];
  char ayl_file[MAX_PATH];
} Config;

enum {
  FILE_MODE_ONCE = 0,
  FILE_MODE_LOOP,
  FILE_MODES_ALL,
};

enum {
  DIR_MODE_ONCE = 0,
  DIR_MODE_LOOP,
  DIR_MODE_SHUFFLE,
  DIR_MODES_ALL,
};

enum {
  VOL_OFF = 0,
  VOL_MIN,
  VOL_MID,
  VOL_MAX,
  VOL_MODES_ALL
};



void config_default()
{
  memset(&Config, 0, sizeof(Config));

  Config.ay_layout = LAY_ABC;
  Config.ay_clock = CLK_SPECTRUM;
  Config.ay_fps = FPS_SPECTRUM;
  Config.scr_bright = 100;
  Config.scr_timeout = 5;
  Config.file_mode = FILE_MODE_ONCE;
  Config.dir_mode = DIR_MODE_ONCE;
  Config.sound_vol = VOL_MIN;

  browser_reset_directory();
}



void config_load()
{

  noInterrupts(); //will crash otherwise, as any Flash access seemingly conflicts with the sigma-delta DAC somehow.

  config_default();

  LittleFS.begin();

  fs::File f = LittleFS.open(CFG_FILENAME, "r");

  if (f)
  {
    if (f.size() == sizeof(Config))
    {
      f.readBytes((char*)&Config, sizeof(Config));
    }

    f.close();
  }

  LittleFS.end();

  interrupts();
}



void config_save()
{
  noInterrupts();

  LittleFS.begin();

  fs::File f = LittleFS.open(CFG_FILENAME, "w");

  if (f)
  {
    f.write((char*)&Config, sizeof(Config));

    f.close();
  }

  LittleFS.end();

  interrupts();
}



void config_screen()
{
  const char* const file_modes[] = {"Once", "Loop"};
  const char* const dir_modes[] = {"Once", "Loop", "Shuffle"};
  const char* const sound_vol_names[] = {"Off", "Min", "Mid", "High"};
  char buf[32];

  bool update = true;

  bool pad_was_down = (pad_state & PAD_ANY) ? true : false;

  while (1)
  {
    if (!(frame_cnt & 7)) update = true;

    if (pad_force_redraw()) update = true;

    if (display_is_active() && update)
    {
      update = false;

      display_clear();

      display_rect(0, 0, 53, 8);
      display_mode(DM_AND);
      display_str_P(4, 1, PSTR("Settings"));
      display_mode(DM_OR);

      int page = Config.cfg_cur < 6 ? 0 : 1;

      sprintf(buf, "%u/2", page + 1);
      display_str(VIEW_WDT - display_str_len(buf), 1, buf);

      if (!page)
      {
        sprintf(buf, "%s", ay_layout_names[Config.ay_layout]);

        display_str_P(6, 10 + 0 * 8, PSTR("Stereo"));
        display_str(VIEW_WDT - display_str_len(buf), 10 + 0 * 8, buf);

        switch (Config.ay_clock)
        {
          case CLK_SPECTRUM: strcpy(buf, "ZX 1.77MHz"); break;
          case CLK_PENTAGON: strcpy(buf, "PEN 1.75MHz"); break;
          case CLK_MSX: strcpy(buf, "MSX 1.78MHz"); break;
          case CLK_CPC: strcpy(buf, "CPC 1.0MHz"); break;
          case CLK_ATARIST: strcpy(buf, "ST 2.0MHz"); break;
        }

        display_str_P(6, 10 + 1 * 8, PSTR("Clock"));
        display_str(VIEW_WDT - display_str_len(buf), 10 + 1 * 8, buf);

        switch (Config.ay_fps)
        {
          case FPS_PENTAGON: strcpy(buf, "48.8Hz"); break;
          case FPS_SPECTRUM: strcpy(buf, "50.0Hz"); break;
          case FPS_NTSC:  strcpy(buf, "60.0Hz"); break;
        }

        display_str_P(6, 10 + 2 * 8, PSTR("FPS"));
        display_str(VIEW_WDT - display_str_len(buf), 10 + 2 * 8, buf);

        sprintf(buf, "%2u%%", Config.scr_bright);
        display_str_P(6, 10 + 3 * 8, PSTR("Scr.brightness"));
        display_str(VIEW_WDT - display_str_len(buf), 10 + 3 * 8, buf);

        if (!Config.scr_timeout)
        {
          strcpy(buf, "Off");
        }
        else
        {
          sprintf(buf, "%2us", Config.scr_timeout);
        }

        display_str_P(6, 10 + 4 * 8, PSTR("Scr.timeout"));
        display_str(128 - display_str_len(buf), 10 + 4 * 8, buf);

        strcpy_P(buf, Config.scr_lock ? PSTR("On") : PSTR("Off"));
        display_str_P(6, 10 + 5 * 8, PSTR("Auto Lock"));
        display_str(128 - display_str_len(buf), 10 + 5 * 8, buf);
      }
      else
      {
        display_str_P(6, 10 + 0 * 8, PSTR("File mode"));
        display_str(VIEW_WDT - display_str_len(file_modes[Config.file_mode]), 10 + 0 * 8, file_modes[Config.file_mode]);

        display_str_P(6, 10 + 1 * 8, PSTR("Dir.mode"));
        display_str(VIEW_WDT - display_str_len(dir_modes[Config.dir_mode]), 10 + 1 * 8, dir_modes[Config.dir_mode]);

        display_str_P(6, 10 + 2 * 8, PSTR("Sounds"));
        display_str(VIEW_WDT - display_str_len(sound_vol_names[Config.sound_vol]), 10 + 2 * 8, sound_vol_names[Config.sound_vol]);

        display_str_P(6, 10 + 3 * 8, PSTR("Reset to default"));

        display_str_P(6, 10 + 4 * 8, PSTR("About"));
      }

      int ccur = Config.cfg_cur < 6 ? Config.cfg_cur : Config.cfg_cur - 6;

      display_char(0, 10 + ccur * 8, !(frame_cnt & 8) ? 0x00 : 0x20);
    }

    display_update();

    wait_frame();

    pad_check();

    if (pad_was_down)
    {
      if (!(pad_state & PAD_ANY)) pad_was_down = false;
    }
    else
    {
      if (pad_state_rep & PAD_REW)
      {
        sound_play(SFX_MOVE);
        update = true;
        frame_cnt = 0;
        --Config.cfg_cur;
        if (Config.cfg_cur < 0) Config.cfg_cur = 10;
      }

      if (pad_state_rep & PAD_FWD)
      {
        sound_play(SFX_MOVE);
        update = true;
        frame_cnt = 0;
        ++Config.cfg_cur;
        if (Config.cfg_cur > 10) Config.cfg_cur = 0;
      }

      if (pad_state_up & PAD_STOP)
      {
        sound_play(SFX_CANCEL);
        break;
      }

      if (pad_state_down & PAD_PLAY)
      {
        sound_play(SFX_SELECT);

        update = true;
        frame_cnt = 0;

        switch (Config.cfg_cur)
        {
          case 0:
            ++Config.ay_layout;
            if (Config.ay_layout >= LAY_ALL) Config.ay_layout = 0;
            break;

          case 1:
            switch (Config.ay_clock)
            {
              case CLK_SPECTRUM: Config.ay_clock = CLK_PENTAGON; break;
              case CLK_PENTAGON: Config.ay_clock = CLK_MSX; break;
              case CLK_MSX: Config.ay_clock = CLK_CPC; break;
              case CLK_CPC: Config.ay_clock = CLK_ATARIST; break;
              case CLK_ATARIST: Config.ay_clock = CLK_SPECTRUM; break;
            }
            break;

          case 2:
            switch (Config.ay_fps)
            {
              case FPS_SPECTRUM: Config.ay_fps = FPS_PENTAGON; break;
              case FPS_PENTAGON: Config.ay_fps = FPS_NTSC; break;
              case FPS_NTSC: Config.ay_fps = FPS_SPECTRUM; break;
            }
            break;

          case 3:
            Config.scr_bright -= 10;
            if (Config.scr_bright < 20) Config.scr_bright = 100;
            display_brightness(Config.scr_bright);
            break;

          case 4:
            switch (Config.scr_timeout)
            {
              case 0: Config.scr_timeout = 1; break;
              case 1: Config.scr_timeout = 3; break;
              case 3: Config.scr_timeout = 5; break;
              case 5: Config.scr_timeout = 10; break;
              case 10: Config.scr_timeout = 20; break;
              case 20: Config.scr_timeout = 30; break;
              case 30: Config.scr_timeout = 60; break;
              case 60: Config.scr_timeout = 90; break;
              case 90: Config.scr_timeout = 0; break;
            }
            break;

          case 5:
            Config.scr_lock ^= 1;
            break;

          case 6:
            ++Config.file_mode;
            if (Config.file_mode >= FILE_MODES_ALL) Config.file_mode = 0;
            break;

          case 7:
            ++Config.dir_mode;
            if (Config.dir_mode >= DIR_MODES_ALL) Config.dir_mode = 0;
            break;

          case 8:
            ++Config.sound_vol;
            if (Config.sound_vol >= VOL_MODES_ALL) Config.sound_vol = 0;
            break;

          case 9:
            if (message_box_P(PSTR("Are you sure?"), MB_YESNO))
            {
              config_default();
              config_save();
              ESP.reset();
            }

            break;

          case 10:
            message_box_P(PSTR(MSG_SPLASH_SCREEN), MB_OK);
            break;
        }
      }
    }
  }

  config_save();
}
