// hway - a hardware AY-3-8910 chiptune music player
// hardware and firmware by shiru8bit (shiru@mail.ru)

// Features:
//
//  - SD card with FAT32 and long file names support
//  - STC, STP, PT1, PT2, PT3, ASC, PSC, SQT support powered via libayfly
//  - Limited AY file format support with beeper and AY music, via z80emu
//  - AYL playlists support
//  - Configurable ABC layouts, chip clocks, frame rates
//  - Fast forward and slowdown (for all formats but AY)
//  - Software lock
//  - Settings and current directory stored in the Flash memory between power ons

// History:
//
//v1.21 08.12.22 - SPI slow and fast implementations, for debug purposes.
//v1.2  11.10.21 - SPI OLED screen support, ESP sound pin moved to RX (warning, loud noises during flashing).
//                 ESP_EEPROM turned to be troublesome, reverted to LittleFS again. Minor fixes.
//v1.11 20.08.21 - A couple of minor fixes here and there. String vars eliminated altogether to avoid potential
//                 memory leaks and fragmentation during lengthy sessions. LittleFS proved to be inefficient for
//                 this use, replaced to ESP_EEPROM.
//v1.1  19.08.21 - Limited AY file format support, complete with Z80 emulation and basic beeper implementation
//                 (not so high quality. ZX Spectrum only). Battery code tweaked up again. Unknown play time
//                 supported (plays indefinitely). Moved from SPIFFS to LittleFS.
//v1.02 04.08.21 - Second AY chip support added as an option, to play TurboSound PT3 modules (6 channels)
//v1.01 27.07.21 - Battery charge measurement and display logic is tweaked up a bit, wasn't working properly at all
//v1.00 30.03.20 - Initial version with all basic functionality
//      08.03.20 - Active development of the initial version started
//      16.01.20 - Project started, based on PT3Play for ESPboy

// Third-party code credits:
//
// brzo_i2c lib by Pascal Kurtansky
// MCP23S17 lib by Majenko Technologies
// libayfly by Deryabin Andrew, this project mostly borrows the format players originally reversed and re-implemented by S.V. Bulba
// z80emu by Lin Ke-Fong, modified in places

#include <ESP8266WiFi.h>
#include <Wire.h>
#include <SPI.h>
#include <MCP23S17.h>
#include <sigma_delta.h>
#include <I2S.h>
#include <LittleFS.h>
#include "SdFat.h"

//#define DEBUG_BATTERY
//#define DEBUG_ALLOW_SERIAL

#define MSG_SPLASH_SCREEN           "HWAY v1.21\nby Shiru 12'22\npowered with\nlibayfly\nand z80emu"
#define MSG_WAIT_SCREEN             "\7Wait"
#define MSG_ERR_NO_CARD             "No SD card"
#define MSG_ERR_CANT_LOAD_FILE      "Can't load file"
#define MSG_ERR_TOO_LARGE           "File is too large"
#define MSG_ERR_UNK_FORMAT          "Unknown file contents"
#define MSG_ERR_CANT_OPEN_DIR       "Can't open dir"
#define MSG_ERR_NOT_ENOUGH_MEMORY   "Not enough memory"
#define MSG_ERR_NO_FILES            "No files"

#define SD_SPI_SPEED      SPI_HALF_SPEED  //just in case there is a need to reduce/increase SD SPI speed

//#ifdef SCREEN_i2c   //i2c is a legacy version, it works just fine but stutters while playing beeper music
#define SCREEN_SPI
#define SCREEN_SPI_FAST

#ifdef SCREEN_i2c
#define OLED_i2c_ADR      0x3c
#define OLED_i2c_SCL      D1
#define OLED_i2c_SDA      D2
#endif

#ifdef SCREEN_SPI_FAST
#define SCREEN_SPI
#endif

#ifdef SCREEN_SPI
#define OLED_SPI_CS       D1
#define OLED_SPI_DC       D2
#define OLED_SPI_RES      D3
#endif

#define MCP_CS            D0
#define SD_CS             D4

#define ESP_SND_PIN       RX    //can't be TX, won't allow ESP to reset

#define TIMER_RATE        44100

#define AY1_MCP_PIN_BDIR  0
#define AY2_MCP_PIN_BDIR  3
#define AY_MCP_PIN_BC1    1
#define AY_MCP_PIN_RESET  2

#define MAX_PATH          260



MCP23S17 mcp(&SPI, MCP_CS, 0);

using namespace sdfat;

SdFat sd_fat;
SdFile sd_dir;
SdFile sd_file;


char lfn[MAX_PATH]; //common array for all lfn operations

volatile uint32_t frame_cnt;
volatile uint32_t frame_div;
volatile uint32_t frame_max;
volatile uint32_t frame_display_div;
volatile uint16_t frame_display_timeout;



enum {
  LAY_ABC = 0,
  LAY_ACB,
  LAY_BAC,
  LAY_BCA,
  LAY_CAB,
  LAY_CBA,
  LAY_ALL
};

const char* ay_layout_names[] = {"ABC", "ACB", "BAC", "BCA", "CAB", "CBA"};

#define CLK_SPECTRUM  	1773400
#define CLK_PENTAGON  	1750000
#define CLK_MSX			    1789772
#define CLK_CPC			    1000000
#define CLK_ATARIST		  2000000

#define FPS_PENTAGON  	48828
#define FPS_SPECTRUM  	50000
#define FPS_NTSC	  	  60000

enum {
  SFX_CANCEL = 0,
  SFX_MOVE,
  SFX_SELECT,
  SFX_STARTUP
};

uint8_t music_data[30 * 1024];
uint32_t music_data_size;

enum {
  FILE_ERR_NONE = 0,
  FILE_ERR_NO_CARD = -1,
  FILE_ERR_TOO_LARGE = -2,
  FILE_ERR_NOT_ENOUGH_MEMORY = -3,
  FILE_ERR_UNK_FORMAT = -4,
  FILE_ERR_OTHER = -5
};

int battery_level();
void wait_frame();
void frame_rate_set(uint32_t fps);
void frame_set_display_timeout(int t);
void display_enable(bool ena);
bool display_is_active();
void sound_play(int id);
bool playlist_open(const char* filename, int cur);
bool playlist_iterate(char* buf, int buf_size);
void playlist_close();
void playlist_absolute_path(char* path, int path_size);
void playlist_file_name(char* path, int path_size);
int AY_PlaySample();
void browser_reset_directory();

#include "res/font.h"
#include "res/logo.h"
#include "res/cancel.h"
#include "res/move.h"
#include "res/select.h"
#include "res/startup.h"

#include "pad.h"
#include "1x06_drv.h"
#include "battery.h"
#include "ui.h"
#include "config.h"
#include "browser.h"
#include "playlist.h"


#include "hwa.h"
#include "ayflyw.h"
#include "sound.h"

#include "ASCPlay.h"
#include "PSCPlay.h"
#include "PT1Play.h"
#include "PT2Play.h"
#include "PT3Play.h"
#include "SQTPlay.h"
#include "STCPlay.h"
#include "STPPlay.h"
#include "AYPlay.h"

#include "specanal.h"
#include "player.h"




void IRAM_ATTR timer_ISR()
{
  sigmaDeltaWrite(0, Sound.dac);

  sound_update();

  ++frame_div;

  if (frame_div >= frame_max)
  {
    frame_div = 0;

    ++frame_cnt;
  }

  ++frame_display_div;

  if (frame_display_div >= TIMER_RATE)
  {
    frame_display_div = 0;

    if (frame_display_timeout > 0)
    {
      --frame_display_timeout;

      if (frame_display_timeout <= 0)
      {
        frame_display_timeout = 0;

        if (display_is_active())
        {
          display_enable(false);

          if (Config.scr_lock) pad_lock(true);
        }
      }
    }
  }
}



void wait_frame()
{
  uint32_t prev = frame_cnt;

  while (frame_cnt == prev) yield();
}



void frame_set_display_timeout(int t)
{
  frame_display_div = 0;
  frame_display_timeout = t ? t : Config.scr_timeout;
}



void frame_rate_set(uint32_t fps)
{
  frame_cnt = 0;
  frame_div = 0;
  frame_max = TIMER_RATE * 1000 / fps;
}



void setup() {

  Serial.begin(115200);

  WiFi.disconnect();
  WiFi.mode(WIFI_OFF);
  WiFi.forceSleepBegin();
  delay(1);

  //messing around trying to resolve SD/MCP occassional conflict, probably not needed

  pinMode(MCP_CS, OUTPUT);
  digitalWrite(MCP_CS, HIGH);
  pinMode(SD_CS, OUTPUT);
  digitalWrite(SD_CS, HIGH);
#ifdef SCREEN_SPI
  pinMode(OLED_SPI_CS, OUTPUT);
  digitalWrite(OLED_SPI_CS, HIGH);
#endif
  pinMode(SCK, OUTPUT);
  digitalWrite(SCK, LOW);

  SPI.begin();

  mcp.begin();

  for (int i = 0; i < 16; ++i)
  {
    if (i >= 4 && i < 8)
    {
      mcp.pinMode(i, INPUT_PULLUP);
    }
    else
    {
      mcp.pinMode(i, OUTPUT);
      mcp.digitalWrite(i, LOW);
    }
  }

  pad_init();

#ifdef SCREEN_i2c
  display_init(OLED_i2c_ADR, OLED_i2c_SCL, OLED_i2c_SDA, font_data, font_width);
#endif
#ifdef SCREEN_SPI
  display_init(OLED_SPI_CS, OLED_SPI_DC, OLED_SPI_RES, font_data, font_width);
#endif

  i2s_begin();

  config_load();

  display_brightness(Config.scr_bright);

  hway_set_clock(Config.ay_clock);
  hway_reset();

  frame_rate_set(Config.ay_fps);
  frame_set_display_timeout(0);

  sound_init();

#ifdef DEBUG_ALLOW_SERIAL
  if ((ESP_SND_PIN == TX) || (ESP_SND_PIN == RX)) pinMode(ESP_SND_PIN, FUNCTION_3);
#endif

  battery_init();

  noInterrupts();

  sigmaDeltaSetup(0, F_CPU / 256);
  sigmaDeltaAttachPin(ESP_SND_PIN);
  sigmaDeltaEnable();

  timer1_attachInterrupt(timer_ISR);
  timer1_enable(TIM_DIV1, TIM_EDGE, TIM_LOOP);
  timer1_write(80000000 / TIMER_RATE);  //timer function considers 80 Mhz even in 160 MHz mode

  interrupts();
}



void test_screen()
{
  hway_reset();
  hway_write(0, 0, 10);
  hway_write(0, 1, 1);
  hway_write(0, 8, 15);
  hway_write(0, 7, 9);

  signed int strip_x = -8;

  while (1)
  {
    display_clear();

    display_mask(2);

    display_rect(strip_x, 0, 8, VIEW_HGT);

    display_mask(0);
    display_rect(0, 0, VIEW_WDT, 8);
    display_mode(DM_AND);

    display_str_P(1, 1, PSTR("HW test"));

    display_mode(DM_OR);

    for (int i = 0; i < 16; ++i) display_str((i & 7) * 10, 16 + i / 8 * 10, mcp.digitalRead(i) ? "1" : "0");

    display_update();

    delay(1000 / 50);

    ++strip_x;

    if (strip_x >= VIEW_WDT) strip_x = -8;
  }
}



void logo_screen()
{
  sound_play(SFX_STARTUP);

  int ptr = 0;

  while (ptr < sizeof(logo_data))
  {
    display_clear();

    memcpy_P(dv_raster, (PGM_P)&logo_data[ptr], sizeof(dv_raster));

    ptr += sizeof(dv_raster);

    display_update();

    delay(1000 / 15);
  }
}



void loop()
{
  //playlist_test();

  pad_check();

  if (pad_state & PAD_STOP) test_screen();

  logo_screen();

  while (1)
  {
    config_save();

    if (strlen(Config.ayl_file) > 0) //playlist mode
    {
      while (1)
      {
        int cur = browser_screen(BROWSE_AYL);

        if (cur < -1) break;
        if (cur < 0)
        {
          config_screen();
          continue;
        }

        while (1)
        {
          config_save();

          if (!playlist_get_entry_full_path(Config.dir_cur, lfn, sizeof(lfn))) break;

          if (!player_screen(lfn)) break;
        }
      }

      Config.dir_cur = Config.dir_cur_prev;
      Config.ayl_file[0] = 0;
    }

    if (browser_screen(BROWSE_DIR) < 0)
    {
      config_screen();
      continue;
    }

    if (!browser_full_path(Config.dir_cur, lfn, sizeof(lfn))) break;

    if (browser_check_ext(lfn) == TYPE_AYL) //enter playlist mode
    {
      strncpy(Config.ayl_file, lfn, sizeof(Config.ayl_file) - 1);

      Config.dir_cur_prev = Config.dir_cur;
      Config.dir_cur = 0;

      continue;
    }

    while (1)
    {
      config_save();

      if (!browser_full_path(Config.dir_cur, lfn, sizeof(lfn))) break;

      if (!player_screen(lfn)) break;
    }
  }
}
